/*
 * dialogs.cpp - routines to handle PhotoPrint's dialogs - namely:
 * Print setup
 * Colour Management setup
 *
 * Copyright (c) 2004-2006 by Alastair M. Robinson
 * Distributed under the terms of the GNU General Public License -
 * see the file named "COPYING" for more details.
 *
 */


#include <iostream>

#include <gutenprint/vars.h>

#include <gtk/gtk.h>
#include <gtk/gtknotebook.h>
#include "pp_printoutput.h"
#include "pp_cms.h"
#include "pp_units.h"
#include "pp_scaling.h"
#include "stpui_widgets/stpui_optionbook.h"
#include "splashscreen.h"
#include "patheditor.h"
#include "imageselector.h"
#include "dialogs.h"
#include "config.h"

using namespace std;


// Colour Management

struct colourmanagementdata
{
	PhotoPrint_State *state;
	GtkWidget *dialog;
};

void ColourManagement_Dialog(GtkWindow *parent,PhotoPrint_State &state)
{
	pp_cms_run_dialog(&state,parent);
}

void Units_Dialog(GtkWindow *parent,PhotoPrint_State &state)
{
	enum Units units=state.GetUnits();
	enum Units newunits=pp_units_run_dialog(units,parent);
	if(units!=newunits)
		state.SetUnits(newunits);
}


void Scaling_Dialog(GtkWindow *parent,PhotoPrint_State &state)
{
	IS_ScalingQuality q=IS_ScalingQuality(state.FindInt("ScalingQuality"));
	q=pp_scaling_run_dialog(parent,q);
	state.SetInt("ScalingQuality",q);
}

//  Print Setup


struct printsetupdata
{
	PhotoPrint_State *state;
	GtkWidget *optionbook;
	GtkWidget *printoutput;
	GtkWidget *dialog;
	stp_vars_t *backupvars;
	char *olddriver;
	char *oldcommand;
	stpui_optionbook_custompage custompage;
};


static void driver_changed(GtkWidget *wid,gpointer data)
{
	cerr << "In driver_changed()" << endl;

	struct printsetupdata *cbd=(struct printsetupdata *)data;
	PrintOutput *po=&cbd->state->printoutput;

	char *driver=po->FindString("Driver");
	cerr << "Driver=" << driver << endl;

	if(!(cbd->state->printer.SetDriver(po->FindString("Driver"))))
	{
		cerr << "Setting driver failed - reverting to default" << endl;
		po->SetString("Driver",DEFAULT_PRINTER_DRIVER);
		cbd->state->printer.SetDriver(po->FindString("Driver"));
		pp_printoutput_refresh(PP_PRINTOUTPUT(cbd->printoutput));
	}

	stpui_optionbook_rebuild(STPUI_OPTIONBOOK(cbd->optionbook));	
}


void PrintSetup_Dialog(GtkWindow *parent,PhotoPrint_State &state)
{
	printsetupdata dialogdata;
	
	dialogdata.state=&state;
	dialogdata.backupvars=stp_vars_create_copy(state.printer.stpvars);
	dialogdata.olddriver=g_strdup(state.printoutput.FindString("Driver"));
	dialogdata.oldcommand=g_strdup(state.printoutput.FindString("Command"));

	dialogdata.dialog=gtk_dialog_new_with_buttons("Printer Setup",
		parent,GtkDialogFlags(0),
		GTK_STOCK_CANCEL,GTK_RESPONSE_CANCEL,
		GTK_STOCK_OK,GTK_RESPONSE_OK,
		NULL);
	gtk_window_set_default_size(GTK_WINDOW(dialogdata.dialog),500,350);

	dialogdata.printoutput=pp_printoutput_new(&state.printoutput);
	g_object_ref(G_OBJECT(dialogdata.printoutput));

	dialogdata.custompage.name="Output";
	dialogdata.custompage.widget=dialogdata.printoutput;
	
	dialogdata.optionbook=stpui_optionbook_new(state.printer.stpvars,&dialogdata.custompage,1);
	gtk_box_pack_start(GTK_BOX(GTK_DIALOG(dialogdata.dialog)->vbox),dialogdata.optionbook,TRUE,TRUE,0);
	gtk_widget_show(dialogdata.optionbook);
	
	g_signal_connect(G_OBJECT(dialogdata.printoutput),"changed",G_CALLBACK(driver_changed),&dialogdata);

	gtk_widget_show(dialogdata.dialog);
	gint result=gtk_dialog_run(GTK_DIALOG(dialogdata.dialog));
	switch(result)
	{
		case GTK_RESPONSE_CANCEL:
			stp_vars_copy(state.printer.stpvars,dialogdata.backupvars);
			state.printoutput.SetString("Driver",dialogdata.olddriver);
			state.printoutput.SetString("Command",dialogdata.oldcommand);
			break;
		case GTK_RESPONSE_OK:
			const char *p1=stp_get_string_parameter(dialogdata.backupvars,"PageSize");
			const char *p2=stp_get_string_parameter(state.printer.stpvars,"PageSize");
			if(strcmp(p1,p2)!=0)
				state.layout->UpdatePageSize();  // Needs to be run every time, but preserve margins.
			break;
	}
	g_object_unref(G_OBJECT(dialogdata.printoutput));
	stp_vars_destroy(dialogdata.backupvars);
	g_free(dialogdata.olddriver);
	g_free(dialogdata.oldcommand);
	gtk_widget_destroy(dialogdata.dialog);
}


// About dialog

void About_Dialog(GtkWindow *parent)
{
	GtkWidget *dialog=gtk_dialog_new_with_buttons("About",
		parent,GtkDialogFlags(0),
		GTK_STOCK_OK,GTK_RESPONSE_OK,
		NULL);

	GdkPixbuf *pb=SplashScreen::GetPixbuf();

	GtkWidget *image=gtk_image_new_from_pixbuf(pb);
	gtk_box_pack_start(GTK_BOX(GTK_DIALOG(dialog)->vbox),image,FALSE,FALSE,0);
	gtk_widget_show(image);
	
	GtkWidget *label=gtk_label_new("PhotoPrint "
	PACKAGE_VERSION
	" - copyright (c) 2004-2006\n"
	"by Alastair M. Robinson (blackfive@fakenhamweb.co.uk)\nDistributed under the terms\nof the GNU General Public Licence.\n"
	"See the file 'COPYING' for more details.");
	gtk_label_set_justify(GTK_LABEL(label),GTK_JUSTIFY_CENTER);
	gtk_box_pack_start(GTK_BOX(GTK_DIALOG(dialog)->vbox),label,TRUE,FALSE,5);
	gtk_widget_show(label);

	gtk_widget_show(dialog);
	gtk_dialog_run(GTK_DIALOG(dialog));

	gtk_widget_destroy(dialog);
	g_object_unref(G_OBJECT(pb));
}


struct PathDialogData
{
	SearchPath sp;
	ImageSelector *is;
	PathEditor *pe;
};


static void	borderpaths_changed(GtkWidget *widget,gpointer user_data)
{
	PathDialogData *dd=(PathDialogData *)user_data;
	patheditor_get_paths(dd->pe,&dd->sp);
	imageselector_refresh(IMAGESELECTOR(dd->is));
}


void Paths_Dialog(GtkWindow *parent,PhotoPrint_State &state)
{
	GtkWidget *dialog;
	dialog=gtk_dialog_new_with_buttons("Set paths...",
		parent,GtkDialogFlags(0),
		GTK_STOCK_CANCEL,GTK_RESPONSE_CANCEL,
		GTK_STOCK_OK,GTK_RESPONSE_OK,
		NULL);
	gtk_window_set_default_size(GTK_WINDOW(dialog),500,350);

	GtkWidget *book=gtk_notebook_new();
	
	gtk_box_pack_start(GTK_BOX(GTK_DIALOG(dialog)->vbox),book,TRUE,TRUE,0);
	gtk_widget_show(book);

	GtkWidget *label;
	GtkWidget *frame;

	frame=gtk_frame_new(NULL);
	gtk_frame_set_shadow_type(GTK_FRAME(frame),GTK_SHADOW_NONE);
	gtk_container_set_border_width(GTK_CONTAINER(frame),4);

	GtkWidget *profilepath=patheditor_new(&state.profilemanager);
	gtk_container_add(GTK_CONTAINER(frame),profilepath);

	gtk_widget_show(profilepath);
	gtk_widget_show(frame);

	label=gtk_label_new("ICC Profiles");
	gtk_notebook_append_page(GTK_NOTEBOOK(book),frame,label);


	PathDialogData dd;
	
	frame=gtk_frame_new(NULL);
	gtk_frame_set_shadow_type(GTK_FRAME(frame),GTK_SHADOW_NONE);
	gtk_container_set_border_width(GTK_CONTAINER(frame),4);
	gtk_widget_show(frame);

	GtkWidget *hbox=gtk_hbox_new(FALSE,0);
	gtk_widget_show(hbox);
	gtk_container_add(GTK_CONTAINER(frame),hbox);

	GtkWidget *borderpath=patheditor_new(&state.bordersearchpath);
	dd.pe=PATHEDITOR(borderpath);
	patheditor_get_paths(dd.pe,&dd.sp);

//	gtk_container_add(GTK_CONTAINER(frame),borderpath);
	gtk_box_pack_start(GTK_BOX(hbox),borderpath,TRUE,TRUE,0);
	gtk_widget_show(borderpath);

	GtkWidget *imagesel=imageselector_new(&dd.sp,false);
	dd.is=IMAGESELECTOR(imagesel);
	g_signal_connect(imagesel,"changed",G_CALLBACK(borderpaths_changed),imagesel);
	gtk_box_pack_start(GTK_BOX(hbox),imagesel,FALSE,FALSE,0);
	gtk_widget_show(imagesel);

	g_signal_connect(borderpath,"changed",G_CALLBACK(borderpaths_changed),&dd);

	label=gtk_label_new("Borders");
	gtk_notebook_append_page(GTK_NOTEBOOK(book),frame,label);

	gtk_widget_show(dialog);

	imageselector_refresh(IMAGESELECTOR(imagesel));

	gint result=gtk_dialog_run(GTK_DIALOG(dialog));
	switch(result)
	{
		case GTK_RESPONSE_CANCEL:
			break;
		case GTK_RESPONSE_OK:
			patheditor_get_paths(PATHEDITOR(profilepath),&state.profilemanager);
			patheditor_get_paths(PATHEDITOR(borderpath),&state.bordersearchpath);
			break;
	}
//	g_object_unref(G_OBJECT(dialogdata.printoutput));
	gtk_widget_destroy(dialog);
}


char *ImageMask_Dialog(GtkWindow *parent,PhotoPrint_State &state,char *oldfn)
{
	char *nfn=oldfn;
	GtkWidget *dialog;
	dialog=gtk_dialog_new_with_buttons("Printer Setup",
		parent,GtkDialogFlags(0),
		GTK_STOCK_CANCEL,GTK_RESPONSE_CANCEL,
		GTK_STOCK_OK,GTK_RESPONSE_OK,
		NULL);
	gtk_window_set_default_size(GTK_WINDOW(dialog),250,350);

	GtkWidget *imagesel=imageselector_new(&state.bordersearchpath,true);
	gtk_box_pack_start(GTK_BOX(GTK_DIALOG(dialog)->vbox),imagesel,TRUE,TRUE,0);
	gtk_widget_show(imagesel);

	gtk_widget_show(dialog);

	imageselector_refresh(IMAGESELECTOR(imagesel));
	imageselector_set_filename(IMAGESELECTOR(imagesel),oldfn);

	gint result=gtk_dialog_run(GTK_DIALOG(dialog));
	switch(result)
	{
		case GTK_RESPONSE_CANCEL:
			break;
		case GTK_RESPONSE_OK:
			if(oldfn)
				free(oldfn);
			nfn=NULL;
			const char *nstr=imageselector_get_filename(IMAGESELECTOR(imagesel));
			if(nstr)
				nfn=strdup(nstr);
			break;
	}
	gtk_widget_destroy(dialog);
	return(nfn);
}
