/*
 * layout.cpp - A superclass for layouts.
 *
 * Copyright (c) 2004 by Alastair M. Robinson
 * Distributed under the terms of the GNU General Public License -
 * see the file named "COPYING" for more details.
 *
 */

#include <iostream>
#include <string.h>

#include "dialogs.h"
#include "generaldialogs.h"
#include "pixbufthumbnail/egg-pixbuf-thumbnail.h"
#include "pixbuf_from_imagesource.h"
#include "rotatepixbuf.h"
#include "maskpixbuf.h"

#include "imagesource/imagesource.h"
#include "imagesource/imagesource_cms.h"
#include "imagesource/imagesource_util.h"
#include "imagesource/imagesource_mask.h"
#include "imagesource/imagesource_rotate.h"
#include "imagesource/imagesource_promote.h"

#include "photoprint_state.h"
#include "progress.h"
#include "util.h"
#include "layoutrectangle.h"

#include "layout.h"

using namespace std;


ConfigTemplate LayoutDB::Template[]=
{
	ConfigTemplate("LayoutType","NUp"),
	ConfigTemplate()
};


Layout_ImageInfo::Layout_ImageInfo(Layout &layout, const char *filename, int page, bool allowcropping, PP_ROTATION rotation)
	: page(page), allowcropping(allowcropping), rotation(rotation), layout(layout),
	maskfilename(NULL), thumbnail(NULL), mask(NULL), transformed(NULL), selected(false)
{
	if(filename[0]=='/')
		this->filename=strdup(filename);
	else
		this->filename=BuildAbsoluteFilename(filename);
	GetThumbnail();
}


Layout_ImageInfo::Layout_ImageInfo(Layout &layout, Layout_ImageInfo *ii, int page, bool allowcropping, PP_ROTATION rotation)
	: page(page), allowcropping(allowcropping), rotation(rotation), layout(layout),
	maskfilename(NULL), thumbnail(NULL), mask(NULL), transformed(NULL), selected(false)
{
	cerr << "Stealing existing image!" << endl;

	if(ii)
	{
		thumbnail=ii->thumbnail;
		ii->thumbnail=NULL;  // We steal the thumbnail from the old image!
		mask=ii->mask;
		ii->mask=NULL;	// And the mask too!
	
		if(ii->filename)
			this->filename=strdup(ii->filename);
		if(ii->maskfilename)
			this->maskfilename=strdup(ii->maskfilename);

		allowcropping=ii->allowcropping;
		rotation=ii->rotation;
	}
}


Layout_ImageInfo::~Layout_ImageInfo()
{
	if(thumbnail)
		g_object_unref(thumbnail);
	if(mask)
		g_object_unref(mask);
	if(transformed)
		g_object_unref(transformed);
	layout.imagelist=g_list_remove(layout.imagelist,this);
	free(filename);
}


void Layout_ImageInfo::DrawThumbnail(GtkWidget *widget,int xpos,int ypos,int width,int height)
{
	GdkPixbuf *thumbnail=GetThumbnail();

	if(transformed)
		g_object_unref(transformed);
	transformed=NULL;
	
	LayoutRectangle r(gdk_pixbuf_get_width(thumbnail),gdk_pixbuf_get_height(thumbnail));
	LayoutRectangle target(xpos,ypos,width,height);

	RectFit *fit=r.Fit(target,allowcropping,rotation,CENTRE,CENTRE);

	GdkPixbuf *tmp;
	switch(fit->rotation)
	{
		case 0:
			transformed=gdk_pixbuf_scale_simple(thumbnail,fit->width,fit->height,GDK_INTERP_NEAREST);
			break;
		case 270:
			tmp=gdk_pixbuf_rotate_simple(thumbnail,GDK_PIXBUF_ROTATE_CLOCKWISE);
			transformed=gdk_pixbuf_scale_simple(tmp,fit->width,fit->height,GDK_INTERP_NEAREST);
			g_object_unref(G_OBJECT(tmp));
			break;
		case 180:
			tmp=gdk_pixbuf_rotate_simple(thumbnail,GDK_PIXBUF_ROTATE_UPSIDEDOWN);
			transformed=gdk_pixbuf_scale_simple(tmp,fit->width,fit->height,GDK_INTERP_NEAREST);
			g_object_unref(G_OBJECT(tmp));
			break;
		case 90:
			tmp=gdk_pixbuf_rotate_simple(thumbnail,GDK_PIXBUF_ROTATE_COUNTERCLOCKWISE);
			transformed=gdk_pixbuf_scale_simple(tmp,fit->width,fit->height,GDK_INTERP_NEAREST);
			g_object_unref(G_OBJECT(tmp));
			break;
	}

	int dw=fit->width-2*fit->xoffset;
	int dh=fit->height-2*fit->yoffset;
	
	if(dw>gdk_pixbuf_get_width(transformed))
	{
		cerr << "DW too high" << endl;
		dw=gdk_pixbuf_get_width(transformed);
	}

	if(dh>gdk_pixbuf_get_height(transformed))
	{
		cerr << "DH too high" << endl;
		dh=gdk_pixbuf_get_height(transformed);
	}

	if(mask)
		maskpixbuf(transformed,fit->xoffset,fit->yoffset,dw,dh,mask);

	gdk_draw_pixbuf(widget->window,NULL,transformed,
		fit->xoffset,fit->yoffset,
		fit->xpos,fit->ypos,
		dw,dh,
		GDK_RGB_DITHER_NONE,0,0);

	delete fit;
}


void Layout_ImageInfo::SetMask(const char *filename)
{
	if(mask)
	{
		g_object_unref(mask);
		mask=NULL;
	}
	if(maskfilename)
	{
		free(maskfilename);
		maskfilename=NULL;
	}
	if(filename)
		maskfilename=strdup(filename);
}


GdkPixbuf *Layout_ImageInfo::GetThumbnail()
{
	GError *err=NULL;

	if(layout.state.batchmode)
		return(NULL);

	if(maskfilename && !mask)
	{
		mask=egg_pixbuf_get_thumbnail_for_file (maskfilename, EGG_PIXBUF_THUMBNAIL_LARGE, &err);
		cerr << "Attempting to load mask from: " << maskfilename << endl;
		if(!mask)
		{
			cerr << "Failed." << endl;
			try
			{
				ImageSource *src=ISLoadImage(maskfilename);
				if(src)
				{
					int w,h;
					w=(src->width*256)/src->height;
					h=256;
					if(w>256)
					{
						w=256;
						h=(src->height*256)/src->width;
					}
					src=ISScaleImageBySize(src,w,h,IS_SCALING_NEARESTNEIGHBOUR);
					mask=pixbuf_from_imagesource(src);
					delete src;
				}
			}
			catch(const char *err)
			{
				cerr << "Error: " << err << endl;
			}	
			if(!mask)
			{
				if(err && err->message)
					cerr << "Error: " << err->message << endl;
				else
					cerr << "Can't get mask thumbnail" << endl;
				free(maskfilename);
				maskfilename=NULL;
			}
		}
	}

	if(thumbnail)
		return(thumbnail);
		
	thumbnail=egg_pixbuf_get_thumbnail_for_file (filename, EGG_PIXBUF_THUMBNAIL_LARGE, &err);

	if(!thumbnail)
	{
		ImageSource *src=ISLoadImage(filename);
		if(src)
		{
			int w,h;
			w=(src->width*256)/src->height;
			h=256;
			if(w>256)
			{
				w=256;
				h=(src->height*256)/src->width;
			}
			src=ISScaleImageBySize(src,w,h,IS_SCALING_NEARESTNEIGHBOUR);
			thumbnail=pixbuf_from_imagesource(src);
			delete src;
		}

		if(!thumbnail)
		{
			if(err && err->message)
				throw err->message;
			else
				throw "Can't get thumbnail";
		}
	}

	return(thumbnail);
}


LayoutRectangle *Layout_ImageInfo::GetBounds()
{
	// Dummy function - override in subclasses!
	LayoutRectangle *result=new LayoutRectangle(0,0,100,100);
	return(result);
}


ImageSource *Layout_ImageInfo::GetImageSource(CMColourDevice target,CMTransformFactory *factory)
{
	ImageSource *result=NULL;
	try
	{
		cerr << "Loading image " << endl;
		ImageSource *is=ISLoadImage(filename);

		if(STRIP_ALPHA(is->type)==IS_TYPE_GREY)
			is=new ImageSource_Promote(is,IS_TYPE_RGB);

		CMSTransform *transform=NULL;

		if(factory)
		{
			if(is->embeddedprofile)
			{
				cerr << "Has embedded profile..." << endl;
				transform=factory->GetTransform(target,is->embeddedprofile);
			}
			else
			{
				cerr << "No embedded profile - using default" << endl;
				transform=factory->GetTransform(target,IS_TYPE(STRIP_ALPHA(is->type)));
			}
	
			if(transform)
				is=new ImageSource_CMS(is,transform);
		}
	
#if 0
		if(layout.state.FindInt("PrinterProfileActive"))
		{
			if(is->embeddedprofile)
			{
				cerr << "Has embedded profile - building transform" << endl;
				char *pf=layout.state.searchpath.Search(layout.state.FindString("PrinterProfile"));
				if(pf)
				{
					CMSProfile *dest=new CMSProfile(pf);
					is=new ImageSource_CMS(is,is->embeddedprofile,dest);
					delete dest;
					free(pf);
				}
				else
					cerr << "Can't find specified printer profile" << endl;
			}
			else
			{
				switch(is->type)
				{
					case IS_TYPE_RGB:
					case IS_TYPE_RGBA:
						if(!layout.defaultrgbtransform)
							layout.GetDefaultRGBTransform();
						if(layout.defaultrgbtransform)
							is=new ImageSource_CMS(is,layout.defaultrgbtransform);
						break;
					case IS_TYPE_CMYK:
					case IS_TYPE_CMYKA:
						if(!layout.defaultcmyktransform)
							layout.GetDefaultCMYKTransform();
						if(layout.defaultcmyktransform)
							is=new ImageSource_CMS(is,layout.defaultcmyktransform);
						break;
					default:
						throw "Only RGB and CMYK images are supported";
				}
			}
		}
#endif
		result=is;
		cerr << "Done" << endl;
	}
	catch(const char *err)
	{
		ErrorMessage_Dialog(err);
	}
	return(result);
}


ImageSource *Layout_ImageInfo::ApplyMask(ImageSource *is)
{
	if(maskfilename)
	{
		ImageSource *mask=ISLoadImage(maskfilename);
		if((is->width>is->height)^(mask->width>mask->height))
		{
			mask=new ImageSource_Rotate(mask,270);
			cerr << "Rotating mask" << endl;
		}
		mask=ISScaleImageBySize(mask,is->width,is->height,IS_SCALING_AUTOMATIC);
		is=new ImageSource_Mask(is,mask);
	}
	return(is);
}


bool Layout_ImageInfo::GetSelected()
{
	return(selected);
}


void Layout_ImageInfo::SetSelected(bool sel)
{
	selected=sel;
}


void Layout_ImageInfo::ToggleSelected()
{
	selected=!selected;
}


const char *Layout_ImageInfo::GetFilename()
{
	return(filename);
}


Layout::Layout(PhotoPrint_State &state,Layout *oldlayout)
	: xoffset(0), yoffset(0), pages(1), currentpage(0), state(state), imagelist(NULL), iterator(NULL)
{
}


Layout::~Layout()
{
	if(imagelist)
	{
		for(unsigned int i=0;i<g_list_length(imagelist);++i)
		{
			GList *element=g_list_nth(imagelist,i);
			Layout_ImageInfo *ii=(Layout_ImageInfo *)element->data;
			delete ii;
		}
	}
	g_list_free(imagelist);
//	if(defaultrgbtransform)
//		delete defaultrgbtransform;
//	if(defaultcmyktransform)
//		delete defaultcmyktransform;
}


void Layout::Clear()
{
	if(imagelist)
	{
		for(unsigned int i=0;i<g_list_length(imagelist);++i)
		{
			GList *element=g_list_nth(imagelist,i);
			Layout_ImageInfo *ii=(Layout_ImageInfo *)element->data;
			delete ii;
		}
	}
	g_list_free(imagelist);
	imagelist=NULL;
	pages=1;
}


// DUMMY FUNCTION - should be overridden by subclasses
int Layout::AddImage(const char *filename,bool allowcropping,PP_ROTATION rotation)
{
	cerr << "AddImage: Dummy function - should be overridden" << endl;
	return(0);
}


// DUMMY FUNCTION - should be overridden by subclasses
void Layout::CopyImage(Layout_ImageInfo *ii)
{
	cerr << "CopyImage: Dummy function - should be overridden" << endl;
}


ImageSource *Layout::GetImageSource(int page,CMColourDevice target,CMTransformFactory *factory)
{
	cerr << "GetImageSource: Dummy function - should be overridden" << endl;
	return(NULL);
}


void Layout::TransferImages(Layout *oldlayout,Progress *p)
{
	if(currentpage>=pages);
	cerr << "Done" << endl;
	if(oldlayout)
	{
		int count=0,img=0;
		Layout_ImageInfo *ii=oldlayout->FirstImage();
		while(ii)
		{
			++count;
			ii=oldlayout->NextImage();
		}
		ii=oldlayout->FirstImage();
		while(ii)
		{
			++img;
			CopyImage(ii);
			ii=oldlayout->NextImage();
			if(p)
			{
				if(!(p->DoProgress(img,count)))
					ii=NULL;
			}
		}
	}
	if(currentpage>=pages)
		currentpage=pages-1;
}


void Layout::DBToLayout(LayoutDB &db)
{
	state.printer.SetDriver(state.printoutput.FindString("Driver"));
}


void Layout::LayoutToDB(LayoutDB &db)
{
}


void Layout::Reflow()
{
}


GtkWidget *Layout::CreateWidget()
{
	return(NULL);
}


void Layout::RefreshWidget(GtkWidget *widget)
{
}


int Layout::GetCurrentPage()
{
	return(currentpage);
}


void Layout::SetCurrentPage(int page)
{
	currentpage=page;
}


void Layout::Print(Progress *p)
{
	state.printer.SetProgress(p);
	CMTransformFactory *factory=state.profilemanager.GetTransformFactory();
	for(int p=0;p<GetPages();++p)
	{
		ImageSource *is=GetImageSource(p,CM_COLOURDEVICE_PRINTER,factory);
		if(is)
		{
			state.printer.Print(is,xoffset,yoffset);
			delete is;
		}
	}
	delete factory;
	state.printer.SetProgress(NULL);
}

void (*Layout::SetUnitFunc())(GtkWidget *wid,enum Units unit)
{
	cerr << "This function should be overridden" << endl;
	return(NULL);
}


void Layout::DrawPreview(GtkWidget *widget,int xpos,int ypos,int width,int height)
{
	Layout_ImageInfo *ii=FirstImage();

	while(ii)
	{
		if(currentpage==ii->page)
		{
			cerr << "Rotation: " << ii->rotation << endl;
			ii->DrawThumbnail(widget,xpos,ypos,width,height);
		}
		ii=NextImage();
	}
}


Layout_ImageInfo *Layout::FirstImage()
{
	iterator=imagelist;
	if(iterator)
		return((Layout_ImageInfo *)iterator->data);
	else
		return(NULL);
}


Layout_ImageInfo *Layout::NextImage()
{
	if(iterator)
		iterator=g_list_next(iterator);
	if(iterator)
		return((Layout_ImageInfo *)iterator->data);
	else
		return(NULL);
}


int Layout::GetPages()
{
	return(pages);
}


Layout_ImageInfo *Layout::FirstSelected()
{
	Layout_ImageInfo *ii=FirstImage();
	while(ii)
	{
		if(ii->GetSelected())
			return(ii);
		ii=NextSelected();
	}
	return(NULL);
}


Layout_ImageInfo *Layout::NextSelected()
{
	Layout_ImageInfo *ii=NextImage();
	while(ii)
	{
		if(ii->GetSelected())
			return(ii);
		ii=NextSelected();
	}
	return(NULL);
}


void Layout::SelectAll()
{
	Layout_ImageInfo *ii=FirstSelected();
	while(ii)
	{
		ii->SetSelected(true);
		ii=NextSelected();
	}
}


void Layout::SelectNone()
{
	Layout_ImageInfo *ii=FirstSelected();
	while(ii)
	{
		ii->SetSelected(false);
		ii=NextSelected();
	}
}


Layout_ImageInfo *Layout::ImageAtCoord(int x,int y)
{
	return(NULL);
}


int Layout::GetCapabilities()
{
	return(0);
}
